from . import BaseActor
from ltr.models.loss.target_regression import DenseIouPred

class FcotV2Actor(BaseActor):
    """Actor for training the first stage of FCOT network, including cls_72 branch and regression branch."""
    def __init__(self, net, objective, device, loss_weight=None):
        super().__init__(net, objective, device)
        if loss_weight is None:
            loss_weight = {'test_clf_72': 100, 'test_init_clf_72': 100, 'test_iter_clf_72': 400,
                           'test_clf_18': 0, 'test_init_clf_18': 0, 'test_iter_clf_18': 0,
                           'reg_72': 1}
        self.loss_weight = loss_weight
        self.dense_iou_pred = DenseIouPred(dim=4)

    def __call__(self, data, gen_iou_map=False):
        """
        args:
            data - The input data.
            gen_iou_map -

        returns:
            loss    - the training loss
            stats  -  dict containing detailed losses
        """
        # Run network
        target_scores_18, target_scores_72, offset_maps = self.net(train_imgs=data['train_images'],
                                                                   test_imgs=data['test_images'],
                                                                   train_bb=data['train_anno'])

        ### Regression offline loss ###
        reg_loss_72 = self.objective['reg_72'](offset_maps, data['ind_72'], data['w2h2_72'], radius=2) * self.loss_weight['reg_72']
        if gen_iou_map:
            pred_iou_map = self.dense_iou_pred(offset_maps, data['ind_72'], data['w2h2_72'], radius=5)     # [h, w]
        else:
            pred_iou_map = None

        ### Classification-72 losses ###
        clf_losses_test_72 = self.objective['test_clf_72'](target_scores_72, data['test_label_72'], data['test_anno']) * self.loss_weight['test_clf_72']

        # Total loss
        loss = clf_losses_test_72 + reg_loss_72


        # Log stats
        stats = {'Loss/total': loss.item(),
                 'Loss/reg_72': reg_loss_72.item(),
                 'Loss/target_clf_72': clf_losses_test_72.item(),
                 # 'Loss/target_clf_18': loss_target_classifier_18.item(),
                }

        return loss, stats, target_scores_72[:,0, ...], data['test_images'][:,0, ...], pred_iou_map

