import torch.nn as nn
import torch
from torch.nn import functional as F
import ltr.models.layers.filter as filter_layer
import ltr.models.layers.activation as activation
from ltr.models.layers.distance import DistanceMap
import math
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt


class LBHingeV2(nn.Module):
    """Loss that uses a 'hinge' on the lower bound.
    This means that for samples with a label value smaller than the threshold, the loss is zero if the prediction is
    also smaller than that threshold.
    args:
        error_matric:  What base loss to use (MSE by default).
        threshold:  Threshold to use for the hinge.
        clip:  Clip the loss if it is above this value.
    """
    def __init__(self, error_metric=nn.MSELoss(reduction='sum'), threshold=None, clip=None):
        super().__init__()
        self.error_metric = error_metric
        self.threshold = threshold if threshold is not None else -100
        self.clip = clip

        # num_dist_bins = 100
        # bin_displacement = 0.1
        # init_gauss_sigma = 0.9
        # mask_act = 'sigmoid'
        # mask_init_factor = 3.0
        # score_act = 'relu'
        # self.feat_stride = 4
        # self.distance_map = DistanceMap(num_dist_bins, bin_displacement)
        # # Distance coordinates
        # d = torch.arange(num_dist_bins, dtype=torch.float32).view(1, -1, 1, 1) * bin_displacement
        # if init_gauss_sigma == 0:
        #     init_gauss = torch.zeros_like(d)
        #     init_gauss[0, 0, 0, 0] = 1
        # else:
        #     init_gauss = torch.exp(-1 / 2 * (d / init_gauss_sigma) ** 2)
        #
        # # Module that predicts the target label function (y in the paper)
        # self.label_map_predictor = nn.Conv2d(num_dist_bins, 1, kernel_size=1, bias=False)
        # self.label_map_predictor.weight.data = init_gauss - init_gauss.min()
        #
        # # Module that predicts the target mask (m in the paper)
        # mask_layers = [nn.Conv2d(num_dist_bins, 1, kernel_size=1, bias=False)]
        # if mask_act == 'sigmoid':
        #     mask_layers.append(nn.Sigmoid())
        #     init_bias = 0.0
        # elif mask_act == 'linear':
        #     init_bias = 0.5
        # else:
        #     raise ValueError('Unknown activation')
        # self.target_mask_predictor = nn.Sequential(*mask_layers)
        # self.target_mask_predictor[0].weight.data = mask_init_factor * torch.tanh(2.0 - d) + init_bias
        #
        # # Module that predicts the residual weights (v in the paper)
        # self.spatial_weight_predictor = nn.Conv2d(num_dist_bins, 1, kernel_size=1, bias=False)
        # self.spatial_weight_predictor.weight.data.fill_(1.0)
        #
        # # The score actvation and its derivative
        # if score_act == 'bentpar':
        #     self.score_activation = activation.BentIdentPar(None)
        #     self.score_activation_deriv = activation.BentIdentParDeriv(None)
        # elif score_act == 'relu':
        #     self.score_activation = activation.LeakyReluPar()
        #     self.score_activation_deriv = activation.LeakyReluParDeriv()
        # else:
        #     raise ValueError('Unknown score activation')
        #
        # for func_name in ['spatial_weight_predictor', 'target_mask_predictor', 'label_map_predictor']:
        #     for p in getattr(self, func_name).parameters():
        #         p.requires_grad_(False)

    def forward(self, prediction, label, target_bb=None):
        negative_mask = (label < self.threshold).float()
        positive_mask = (1.0 - negative_mask)
        # print("label shape: {}, prediction shape: {}, positive_mask shape: {}".format(label.shape, prediction.shape, positive_mask.shape))

        ### visualization
        # bb = target_bb.view(-1, 4).clone()[0] / 4
        # center = (bb[0:2] + bb[2:4] / 2).int()
        # # bb = bb.cpu().detach().numpy()
        # label_pos = positive_mask * label
        # save_id = np.random.randint(1, 10000)
        # # pred = prediction[0,0,bb[1]:bb[1]+8, bb[0]:bb[0]+8].cpu().detach().numpy()
        # # gt = label_pos[0,0,bb[1]:bb[1]+8, bb[0]:bb[0]+8].cpu().detach().numpy()
        # pred = prediction[0, 0, center[1]-4 : center[1] + 4, center[0]-4 : center[0] + 4].cpu().detach().numpy()
        # gt = label_pos[0, 0, center[1]-4:center[1] + 4, center[0]-4:center[0] + 4].cpu().detach().numpy()
        # print("pred shape: {}".format(pred.shape))
        # sns.heatmap(pred, annot = True)
        # # s1 = p1.get_figure()
        # plt.savefig("{}_pred.png".format(save_id))
        # plt.close('all')
        # # s1.savefig("{}_pred.png".format(save_id))
        #
        # sns.reset_defaults()
        # sns.heatmap(gt, annot = True)
        # # s2 = p2.get_figure()
        # plt.savefig("{}_gt.png".format(save_id))
        # plt.close('all')
        # # s2.savefig("{}_gt.png".format(save_id))

        ### vis end

        prediction = positive_mask * prediction
        loss = self.error_metric(prediction, positive_mask * label)

        # num_images = prediction.size(0)
        # num_sequences = prediction.size(1)
        # filter_sz = (4, 4)
        # dmap_offset = (torch.Tensor(filter_sz).to(target_bb.device) % 2) / 2.0
        # center = ((target_bb[..., :2] + target_bb[..., 2:] / 2) / self.feat_stride).view(-1, 2).flip((1,)) - dmap_offset
        # # print("bb: {}, center: {}".format(bb, center))
        # output_sz = (72 + (4 + 1) % 2, 72 + (4 + 1) % 2)
        # dist_map = self.distance_map(center, output_sz).to(target_bb.device)
        #
        # self.label_map_predictor.to(target_bb.device)
        # self.target_mask_predictor.to(target_bb.device)
        # self.spatial_weight_predictor.to(target_bb.device)
        #
        # label_map = self.label_map_predictor(dist_map).view(num_images, num_sequences, *dist_map.shape[-2:])
        # target_mask = self.target_mask_predictor(dist_map).view(num_images, num_sequences, *dist_map.shape[-2:])  # m
        # spatial_weight = self.spatial_weight_predictor(dist_map).view(num_images, num_sequences,
        #                                                               *dist_map.shape[-2:])  # v
        #
        # scores_act = self.score_activation(prediction, target_mask)
        # loss = self.error_metric(prediction, label)
        # loss = self.error_metric((target_mask * prediction + (1 - target_mask) * F.relu(prediction)), label)

        if self.clip is not None:
            loss = torch.min(loss, torch.tensor([self.clip], device=loss.device))
        return loss